/***************************************************************************************** 	
		Programa PAINEL_MSG_BY_ARNE.C

		Autor: Márcio José Soares
		Data: 09/10/2004
      Rev:  13/08/2010

		Microcontrolador: PIC16F877 Microchip
		Frequência de clock: 4 MHZ (1 MIPS)

		Compilador: Hitec PICC Lite V9.65PL1 (Linux)
		IDE: Geany - Ubuntu 9.04
		Gravador: McFlash com picp (Linux)

		Objetivo: Escrever mensagens em um painel
		montado com 12 matrizes de LED's 7 x 5
      Total de 420 LED's

		Necessário placa painel_msg_arne.SCH

*****************************************************************************************/

/*****************************************************************************************
//	Inclusões independentes do módulo
//***************************************************************************************/

#include	<pic.h>                 //configura PIC
#include <htc.h>                 //funcoes genericas - delays
#include "delay.c"

#define XTAL 4000000

/*****************************************************************************************
//	Definições do módulo
//***************************************************************************************/
#define porta_linha		PORTD				//porta de dados para linha
#define porta_controle PORTA				//porta de controle (74HC154)
#define CONT 20								//contador para mostrar frames
#define TEMPO 25								//tempo em microsegundos para cada coluna ligada
#define NRCAR 37								//número de caracteres na matriz
#define TAMCAR	5								//tamanho do caracter (5 colunas)
#define TAM (NRCAR*TAMCAR)					//tamanho total da matriz de caracteres
#define MAXBUF	24								//tamanho máximo do buffer de caracteres a escrever

#define COL1	RC0							//bit para ligar colunas
#define COL2	RC1													
#define COL3	RC2
#define COL4	RC3

#define ENTER 0x0D
#define LF		0x0A

/*****************************************************************************************
//	Inclusões que dependem de pré-definições do módulo
//***************************************************************************************/
#include "matriz.c"							//arquivo com os caracteres possíveis
#include "serial.c"							//arquivo para trabalhar com USART

/*****************************************************************************************
//	Variáveis globais do módulo
//****************************************************************************************/
volatile bit RECEBE_DADOS;					//informa inicio ou fim da recepção dos dados						


/*****************************************************************************************
//	Função liga_col - insere um deterninado valor nos 4 primeiros bits do port C
//	Valor = coluna a ser ligada através do 74HC154
//	recebe:
//		número da coluna
//	retorna:
//		nada
//****************************************************************************************/
void liga_col(int col){
		
		COL1=0;								//desliga colunas antes de continuar
		COL2=0;
		COL3=0;
		COL4=0;

		switch (col){						//testa valor recebido
			case 0:
				COL1=0;
				COL2=0;
				COL3=0;
				COL4=0;
				break;
			case 1:
				COL1=1;
				break;
			case 2:
				COL2=1;
				break;
			case 3:
				COL1=1;
				COL2=1;
				break;
			case 4:
				COL3=1;
				break;
			case 5:
				COL1=1;
				COL3=1;
				break;
			case 6:
				COL2=1;
				COL3=1;
				break;
			case 7:
				COL1=1;
				COL2=1;
				COL3=1;
				break;
			case 8:
				COL4=1;
				break;
			case 9:
				COL1=1;
				COL4=1;
				break;
			case 10:
				COL2=1;
				COL4=1;
				break;
			case 11:
				COL1=1;
				COL2=1;
				COL4=1;
				break;
			case 12:
				COL3=1;
				COL4=1;
				break;
			case 13:
				COL1=1;
				COL3=1;
				COL4=1;
				break;
			case 14:
				COL2=1;
				COL3=1;
				COL4=1;
				break;
			case 15:
				COL1=1;
				COL2=1;
				COL3=1;
				COL4=1;
				break;
			default:									//valor recebido fora da faixa (0 a 15)
				COL1=0;
				COL2=0;
				COL3=0;
				COL4=0;
		}
}

/*****************************************************************************************
//	Função liga_74154 - determina qual 74HC154 deve ser ligado de acordo com a coluna passada
//	trabalha valor da coluna passada para ficar dentro da faixa do 74154 - 0 a 15 
//	recebe:
//		número da coluna
//	retorna:
//		nada
//****************************************************************************************/
void liga_74154(int col){

		int val_col;											//valor a ser inserido em porta_controle
		unsigned char new_col;								//variável auxiliar para calcular qual
																	//coluna será ligada, na entrada do 74HC154
		
		if (col <= 15){					  					//coluna dentro do primeiro 74HC154
				val_col=0x0E;									//liga primeiro 74HC154
				new_col=col;
		}
		else if ((col >= 16) && (col < 32)){			//coluna dentro do segundo 74HC154
				val_col=0x0D;									//liga segundo 74HC154
				new_col=col-16;
		}
		else if ((col >= 32) && (col < 48)){ 			//coluna dentro do terceiro 74HC154
				val_col=0x0B;									//liga terceiro 74HC154
				new_col=col-32;
		}
		else if ((col >= 48) && (col < 60)){ 			//coluna dentro do quarto 74HC154	
				val_col=0x07;									//liga quarto 74HC154
				new_col=col-48;
		}
		else{
				val_col=0x0F;									//valor fora. Deslisga todos os 74HC154
				new_col=0x00;
		}

		liga_col(new_col);									//põe valor da coluna no port C
		porta_controle=val_col;								//liga 74HC154
}

/*****************************************************************************************
// Função escreve - escreve um caracter da tabela no display
// sentido para escrever: esquerda para direita	
// recebe:
//		- posição inicial do caracter na tabela principal
//		- coluna inicial onde se deseja grafar o caracter
//	retorna:
//		nada
//****************************************************************************************/
void escreve(char carac, int coluna){
		
		int a,b,c;
		b=0;
		c=0;

		for(a=0;a<5;a++){				  					//contador para coluna do caracter
			porta_linha=linha[a+(carac*5)];			//pega valor para por na linha
			c=coluna-b;	  									//liga coluna correspondente
			liga_74154(c);									//verifica qual coluna deve ligar e liga
			DelayUs(TEMPO);				  				//espera em us
			porta_controle = 0x0F;						//desliga todos 74HC154
			++b;												//incrementa b
		}
}

/*****************************************************************************************
// Função principal (main)
//	recebe:
//		nada
//	retorna:
//		nada:
//****************************************************************************************/
void main (void){
		int a, b, c, d;										//variáveis auxiliares para contadores, etc
		int total, total2;									//total de caracteres a demonstrar no display
		int qframes;											//quantidade de frames a mostrar
		char rec[MAXBUF], rec2[MAXBUF];					//buffer para dados do display
		char carac;												//recebe caracteres
		short int msgrec;										//avisa se mensagem recebida
			
	
		TRISA = 0x00;											//porta A - todos os pinos como saída	
		TRISC = 0x80;											//porta C - todos os pinos como saida, exceto RC7
		TRISD = 0X00;											//porta D - todos os pinos como saída
	
		porta_linha = 0x00;									//zera linha
		porta_controle = 0x0F;		  						//desliga todos 74HC154

		init_comms();											//inicializa USART
		
		for (a=0;a<2;a++)										//aguarda 1 segundo
			DelayMs(250);

		putch(ENTER);
		putch (LF);
		printf("Painel de mensagens -> operando!!!");
		
		COL1=0;													//zera colunas
		COL2=0;
		COL3=0;
		COL4=0;
		
		total2=0;
		RECEBE_DADOS=0;										//ainda não recebeu nada
		msgrec=0;												//se nada foi recebido
																	//escreve default

		//quantidade de caracteres na palavra (ou frase) a escrever
		//durante a inicialização
		total=15;														

		//caracteres a escrever no default

	   rec[0]=25;	//P									
		rec[1]=10;  //A
		rec[2]=18;	//I
		rec[3]=23;	//N
		rec[4]=14;	//E
		rec[5]=21;	//L

		rec[6]=36;	//espaço

		rec[7]=22;  //M
		rec[8]=28;	//S
		rec[9]=16;	//G

		rec[10]=36;	//espaço

		rec[11]=10;	//A									
		rec[12]=27; //R
		rec[13]=23;	//N
		rec[14]=14;	//E


		while(1){

			if (msgrec){															//mensagem recebida
				for (a=0;a<total2+1;a++){										//transfere valores
					if ((rec2[a]>=0x30) && (rec2[a]<=0x39))				//caracter entre 0 e 9
						b=48;
					else if ((rec2[a]>=0x41) && (rec2[a]<=0x5A))			//caracter entre A e Z maisculo
						b=55;															//valor para adaptação na tabela
					else if ((rec2[a]>=0x61) && (rec2[a]<=0x7A))			//caracter entre a e z minusculo
						b=87;
					else{																//caracter invalido, insere espaço
						rec2[a]=36;									
						b=0;
					}
					rec[a]=rec2[a]-b;												//adapta ascii e atualiza tabela a ser mostrada
				}
					total=total2;
					msgrec=0;
			}

			qframes=CONT/total;													//calcula quantos frames
			if ((qframes < 2) && (total <=11))								//não permite menos que dois frames
					qframes=2;

			if (qframes <=0)														//não permite nada menor que 1
					qframes=1;

			for(a=0;a<60+(total*TAMCAR);a++)									//contador para o total de colunas
	  		for(b=0;b<(qframes);b++)											//contador para mostrar frames
					for (c=0;c<total;c++){
						escreve(rec[c],a-(c*5));								//escreve caracteres
						if (RCIF){													//tem dado na porta
								carac=getch();										//recebe caracter
								switch (carac){
									case 0x0D:										//analisa se ENTER
										putch(ENTER);								//devolve mensagem recebida
										putch(LF);
										if (total2 > 0){							//tem mensagem
											printf("Mensagem recebida -> ");
											for(d=0;d<total2;d++)
												putch(rec2[d]);
											RECEBE_DADOS=0;
											msgrec=1;
										}
										else{
											printf("Mensagem vazia!!! Mantendo a antiga!!!");
											RECEBE_DADOS=0;
											msgrec=0;
										}
										break;
									case 0x08:										//analisa BACK SPACE
										putch(carac);
										--total2;									//tira um
										if (total2 < 0)							//impede que contador seja menor que zero
											total2=0;
										break;
									case 0xFE:										//analisa pedido de entrada para mensagem
										RECEBE_DADOS=1;
										total2=0;
										putch(ENTER);
										putch(LF);
										printf("\nEnvie a mensagem->");
										break;
									default:											//recebe mensagem
										if (RECEBE_DADOS){
											if (total2==MAXBUF){					//analisa tamanho da mensagem
												putch(ENTER);
												putch(LF);
												printf("Buffer cheio!!! Tecle ENTER!!!");
												putch(ENTER);
												putch(LF);
										  }
											else{
										  	putch(carac);							//eco do caracter
												rec2[total2]=carac;				//guarda na matriz
												++total2;
											}
										}
								}										
						}
					}
		}
}
